#include <io.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <signal.h>
#ifdef __EMX__
#include <malloc.h>
#else
#include <alloc.h>
#endif
#ifdef __WATCOMC__
#include <direct.h>
#else
#include <dirent.h>
#endif
#ifdef __OS2__
#define INCL_DOSPROCESS
#include <os2.h>
#endif

#ifndef TRUE
#define FALSE 0
#define TRUE 1
#endif

#pragma pack(1)

#define INC_FE_TYPES
#define INC_FE_DATETYPE
#define INC_FE_BAMPROCS
#include "fecfg146.h"

#define MAXLOCKS  60*20  /* waiting 20 mins for febusy */

#ifdef __EMX__
#define mkdir(d)   mkdir(d, 0755)
#endif

char cfgname[256], febusy[256]="";
char *cfg;
CONFIG *basecfg;
#ifdef INCORE
Node *nodes;
Area *areas;
#else
char newcfgname[256];
int  nhcfg;
#endif
SysAddress *myaka;
Packers *packers;
GroupDefaults *grpdefs;
int hcfg, hbsy=-1;
Address zeroaddr={0, 0, 0, 0};
Address allnodes={-1, -1, -1, -1};

void rmbusy(void)
{
  if (febusy[0]==0) return;
  if (hcfg!=-1)
  { close(hcfg);
    hcfg=-1;
  }
  if (hbsy!=-1)
  { close(hbsy);
    hbsy=-1;
    unlink(febusy);
  }
  febusy[0]=0;
}

void cbreak(int signo)
{
  signo++; /* to satisfy compiler */
  printf("\nProgram terminated\n");
  rmbusy();
  exit(4);
}

void doenv(char *str)
{
  /* expand "$[VAR]" to environment variable value */
  char *p, *p1, *var;
  int  i;

  p=str;
  while ((p=strpbrk(p, "$["))!=NULL)
  {
    p1=strchr(p, ']');
    if (p1==NULL) return;
    *p1++='\0';
    var=getenv(p+2);
    if (var==NULL)
    { strcpy(p, p1);
      continue;
    }
    if (strlen(var)<p1-p)
    { strcpy(p, var);
      strcat(p, p1);
      p+=strlen(var);
      continue;
    }
    else if (strlen(var)==p1-p)
    { strncpy(p, var, p1-p);
      p+=strlen(var);
      continue;
    }
    for (i=strlen(p1); i>=0; i--)
      p1[i+strlen(var)-(p1-p)]=p1[i];
    strncpy(p, var, strlen(var));
    p+=strlen(var);
  }
}

#ifdef __MSDOS__

#ifndef __HUGE__
#error This program works only in huge memory model!
#endif

long lread(int fd, void *to, long skoko)
{ long readed=0;
  int  r;

  while (skoko)
  { r=read(fd, to, (skoko>0x4000) ? 0x4000 : (int)skoko);
    if (r==-1) return -1;
    if (r==0) return readed;
    readed+=r;
    to=(char *)to+r;
    skoko-=r;
  }
  return readed;
}

long lwrite(int fd, void *from, long skoko)
{ long wrote=0;
  int  r;

  while (skoko)
  { r=write(fd, from, (skoko>0x4000) ? 0x4000 : (int)skoko);
    if (r==-1) return -1;
    if (r==0) return wrote;
    wrote+=r;
    from=(char *)from+r;
    skoko-=r;
  }
  return wrote;
}

#endif

int readcfg(int ro)
{
  DIR *d;
  struct dirent *df;
  long lcfg;
  char *p;
  char flagsdir[256];
  ExtensionHeader *exthdr;
  int  i;

  /****************** Read fastecho cfg ***************/
  for (i=0; i<10; i++)
  { hcfg=open(cfgname, O_BINARY|O_RDWR);
    if (hcfg==-1)
    { if (errno==EACCES ||
          errno==EAGAIN ||
          errno==EBUSY ||
          errno==ENOENT)
      { if (i==0)
          fprintf(stderr, "Can't open %s: %s (another copy running?), waiting...\n",
                  cfgname, strerror(errno));
        sleep(1);
        continue;
      }
      fprintf(stderr, "Can't open %s: %s!\n", cfgname, strerror(errno));
      return 1;
    }
    else
      break;
  }
  if (i==10)
  { fprintf(stderr, "Can't open %s, exiting\n", cfgname);
    return 1;
  }
  lcfg=filelength(hcfg);
  if (lcfg<sizeof(*basecfg))
  {
    fprintf(stderr, "Incorrect config!\n");
    close(hcfg);
    return 2;
  }
#ifdef INCORE
#ifdef __MSDOS__
  cfg=farmalloc(lcfg);
#else
  cfg=malloc(lcfg);
#endif
  if (cfg==NULL)
  { close(hcfg);
    fprintf(stderr, "Not enough memory!\n");
    return 2;
  }
#ifndef __MSDOS__
  if (read(hcfg, cfg, (int)lcfg)!=lcfg)
#else
  if (lread(hcfg, cfg, lcfg)!=lcfg)
#endif
  { fprintf(stderr, "Can't read %s: %s!\n", cfgname, strerror(errno));
    close(hcfg);
    return 1;
  }
  basecfg=(CONFIG *)cfg;

#if 0
  { ExtensionHeader *exthdr;
    int offset;
    for (offset=sizeof(*basecfg);
         offset < cfg+sizeof(*basecfg)+basecfg->offset;
         offset += exthdr->offset + sizeof(*exthdr))
    {
      exthdr = (ExtensionHeader *)(cfg + offset);
      printf("ExtHdr: type=%u, offset = %u (0x%04x)\n", exthdr->type, offset, offset);
    }
  }
#endif

#else
  basecfg=malloc(sizeof(CONFIG));
  if (basecfg==NULL)
  { close(hcfg);
    fprintf(stderr, "Not enough memory!\n");
    return 2;
  }
  cfg=(char *)basecfg;
  if (read(hcfg, cfg, sizeof(CONFIG))!=sizeof(CONFIG))
  { fprintf(stderr, "Can't read %s: %s!\n", cfgname, strerror(errno));
    close(hcfg);
    return 1;
  }
#endif
  if (basecfg->revision!=6)
  { close(hcfg);
    fprintf(stderr, "Incorrect config revision!\n");
    return 2;
  }
  if (lcfg!=sizeof(*basecfg)+basecfg->offset+
      (long)basecfg->NodeCnt*basecfg->NodeRecSize+
      (long)basecfg->AreaCnt*basecfg->AreaRecSize)
  {
    close(hcfg);
    fprintf(stderr, "Incorrect config length (expected %ld)!\n",
      sizeof(*basecfg)+basecfg->offset+
      (long)basecfg->NodeCnt*basecfg->NodeRecSize+
      (long)basecfg->AreaCnt*basecfg->AreaRecSize);
    return 2;
  }
#ifdef INCORE
  nodes=(Node *)(cfg+sizeof(*basecfg)+basecfg->offset);
  areas=(Area *)(cfg+sizeof(*basecfg)+basecfg->offset+
                 basecfg->NodeCnt*basecfg->NodeRecSize);
#endif
  if (ro)
  { /* complete unlock config, don't set/check febusy.* */
#ifdef INCODE
    close(hcfg);
    hcfg=-1;
#endif
    goto readext;
  }
  /* set semaphore */
  strcpy(febusy, basecfg->SemaphorePath);
  if (febusy[0])
  { doenv(febusy);
    if (febusy[strlen(febusy)-1]!='\\')
      strcat(febusy, "\\");
    strcat(febusy, "FEBUSY.");
    if ((p=getenv("TASK"))!=NULL)
      strcat(febusy, p);
    else
      strcat(febusy, "0");
    for (i=0; i<MAXLOCKS; i++, sleep(1))
    { hbsy=open(febusy, O_BINARY|O_RDWR|O_CREAT|O_EXCL, S_IREAD|S_IWRITE);
      if (hbsy==-1)
      { if (hcfg!=-1)
        { close(hcfg);
          hcfg=-1;
        }
        if (i==0) fprintf(stderr, "Fastecho busy, waiting...\n");
        continue;
      }
      if (i==0) atexit(rmbusy);
      /* check for other semaphores FEBUSY.* */
      strcpy(flagsdir, febusy);
      p=strrchr(flagsdir, '\\');
      if (p==NULL) p=flagsdir;
      else p++;
      strcat(p, ".");
      if ((d=opendir(flagsdir)) != NULL)
      {
        p=strrchr(febusy, '\\');
        if (p==NULL) p=febusy;
        else p++;
        while ((df=readdir(d))!=NULL)
        {
          if (stricmp(df->d_name, p)==0) continue; /* our */
          if (strnicmp(df->d_name, "febusy.", 7)==0)
          { if (hcfg!=-1)
            { close(hcfg);
              hcfg=-1;
            }
            close(hbsy);
            hbsy=-1;
            unlink(febusy);
            if (i==0) fprintf(stderr, "Fastecho busy, waiting...\n");
            break;
          }
        }
        closedir(d);
      }
      if (hbsy!=-1) break;
    }
    if (hbsy==-1)
    { fprintf(stderr, "Can't set febusy semaphore, exiting\n");
      return 244;
    }
    if (hcfg==-1)
    { /* closed while waiting for febusy */
      hcfg=open(cfgname, O_BINARY|O_RDWR);
      if (hcfg==-1)
      { fprintf(stderr, "Can't open %s: %s!\n", cfgname, strerror(errno));
        rmbusy();
        return 1;
      }
    }
  }
readext:
  /* find akas */
  myaka=NULL;
#ifdef INCORE
  for (exthdr=(ExtensionHeader *)(cfg+sizeof(CONFIG));
       (char *)exthdr-cfg<basecfg->offset;
       exthdr=(ExtensionHeader *)((char *)exthdr+exthdr->offset+sizeof(*exthdr)))
  { if (exthdr->type==EH_AKAS)
      myaka=(SysAddress *)((char *)exthdr+sizeof(*exthdr));
#ifdef __MSDOS__
    else if (exthdr->type==EH_PACKERS)
#else
    else if (exthdr->type==EH_PACKERS2)
#endif
      packers=(Packers *)((char *)exthdr+sizeof(*exthdr));
    else if (exthdr->type==EH_GRPDEFAULTS)
      grpdefs=(GroupDefaults *)((char *)exthdr+sizeof(*exthdr));
  }
#else
  exthdr=malloc(sizeof(ExtensionHeader));
  if (exthdr==NULL)
nocore:
  { fprintf(stderr, "Not enough memory!\n");
    rmbusy();
    return 2;
  }
  while (tell(hcfg)<basecfg->offset)
  { read(hcfg, exthdr, sizeof(*exthdr));
    if (exthdr->type==EH_AKAS)
    { myaka=malloc(exthdr->offset);
      if (myaka==NULL) goto nocore;
      read(hcfg, myaka, exthdr->offset);
    }
#ifdef __MSDOS__
    else if (exthdr->type==EH_PACKERS)
#else
    else if (exthdr->type==EH_PACKERS2)
#endif
    { packers=malloc(exthdr->offset);
      if (packers==NULL) goto nocore;
      read(hcfg, packers, exthdr->offset);
    }
    else if (exthdr->type==EH_GRPDEFAULTS)
    { grpdefs=malloc(exthdr->offset);
      if (grpdefs==NULL) goto nocore;
      read(hcfg, grpdefs, exthdr->offset);
    }
    else
      lseek(hcfg, exthdr->offset, SEEK_CUR);
  }
  free(exthdr);
#endif
  if (myaka==NULL)
  { fprintf(stderr, "Incorrect config!\n");
    rmbusy();
    return 2;
  }
  return 0;
}

int lrename(char *oldname, char *newname)
{
  int i=10; /* 10 tries to rename if config locked */

  while (i--)
  { if (rename(oldname, newname)==0)
      return 0;
    if (errno!=EAGAIN && errno!=EACCES && errno!=EBUSY)
      return -1;
    if (i==0)
      fprintf(stderr, "Config locked, waiting...\n");
    sleep(1);
  }
  fprintf(stderr, "Can't rename %s to %s: %s!\n", oldname, newname,
          strerror(errno));
  return -1;
}

#ifdef INCORE
int writecfg(void)
{
  char bakcfg[256];
  long lcfg;

  if (hcfg!=-1)
  { close(hcfg);
    hcfg=-1;
  }
  strcpy(bakcfg, cfgname);
  strcpy(bakcfg+strlen(bakcfg)-3, "bak");
  if (access(bakcfg, 0)==0) unlink(bakcfg);
  if (lrename(cfgname, bakcfg))
  { fprintf(stderr, "Can't rename %s to .bak: %s!\n", cfgname, strerror(errno));
    rmbusy();
    return 2;
  }
  hcfg=open(cfgname, O_BINARY|O_CREAT|O_RDWR|O_EXCL, S_IREAD|S_IWRITE);
  if (hcfg==-1)
  { fprintf(stderr, "Can't create %s: %s!\n", cfgname, strerror(errno));
    lrename(bakcfg, cfgname);
    rmbusy();
    return 2;
  }
  lcfg=sizeof(*basecfg)+basecfg->offset+
       (long)basecfg->NodeCnt*basecfg->NodeRecSize+
       (long)basecfg->AreaCnt*sizeof(Area);
#ifndef __MSDOS__
  if (write(hcfg, cfg, (int)lcfg)!=lcfg)
#else
  if (lwrite(hcfg, cfg, lcfg)!=lcfg)
#endif
  { fprintf(stderr, "Can't write to %s: %s!\n", cfgname, strerror(errno));
    close(hcfg); hcfg=-1;
    unlink(cfgname);
    lrename(bakcfg, cfgname);
    rmbusy();
    return 2;
  }
  rmbusy();
  return 0;
}

#else

int copybytes(int from, int to, long skoko)
{
  static char buf[16384];

  while (skoko>=sizeof(buf))
  { read(from, buf, sizeof(buf));
    if (write(to, buf, sizeof(buf))!=sizeof(buf))
    {
errwrite:
      fprintf(stderr, "Can't write to %s: %s!\n", newcfgname, strerror(errno));
      close(to);
      unlink(newcfgname);
      rmbusy();
      return -1;
    }
    skoko-=sizeof(buf);
  }
  if (skoko)
  { read(from, buf, (int)skoko);
    if (write(to, buf, (int)skoko)!=skoko)
      goto errwrite;
  }
  return 0;
}

int newcfg(void)
{
  int nhcfg;

  strcpy(newcfgname, cfgname);
  strcpy(newcfgname+strlen(newcfgname)-3, "new");
  if (access(newcfgname, 0)==0) unlink(newcfgname);
  nhcfg=open(newcfgname, O_BINARY|O_RDWR|O_CREAT|O_EXCL, S_IREAD|S_IWRITE);
  if (nhcfg==-1)
    fprintf(stderr, "Can't create %s: %s!\n", newcfgname, strerror(errno));
  return nhcfg;
}

int rencfg(int new, int old)
{
  static char bakcfg[256];

  close(new);
  close(old);
  hcfg=-1;
  strcpy(bakcfg, cfgname);
  strcpy(bakcfg+strlen(bakcfg)-3, "bak");
  if (access(bakcfg, 0)==0) unlink(bakcfg);
  if (lrename(cfgname, bakcfg))
  { fprintf(stderr, "Can't rename %s to .bak: %s!\n", cfgname, strerror(errno));
    unlink(newcfgname);
    rmbusy();
    return 2;
  }
  if (lrename(newcfgname, cfgname))
  { fprintf(stderr, "Can't rename %s to .cfg: %s!\n", newcfgname, strerror(errno));
    unlink(newcfgname);
    lrename(bakcfg, cfgname);
    rmbusy();
    return 2;
  }
  rmbusy();
  return 0;
}

#endif

int readgroup(char c)
{
  byte group;
  if (isalpha((unsigned char)c))
    group=c-'A';
  else if (isdigit((unsigned char)c))
    group=c-'0'+'Z'-'A';
  else
    return -1;
  if (group>=32)
    return -1;
  return group;
}

dword readgroups(char *sgroups)
{ dword groups=0;
  byte group;

  for (;*sgroups; sgroups++)
  { if ((group=readgroup(toupper(*sgroups)))==(byte)-1)
    { fprintf(stderr, "Incorrect group %c ignored\n", *sgroups);
      continue;
    }
    AddBam((byte *)&groups, group);
  }
  return groups;
}

int readattr(char *attr, byte *status, byte *direct)
{
  int  r=0;
  status=direct=0;

  for (; *attr; attr++)
  { switch(toupper(*attr))
    { case 'C':  *status=2; break;
      case 'H':  *status=1; break;
      case 'N':  *status=0; break;
      case 'D':  *direct=1; break;
      default:   r=1; break;
    }
  }
  return r;
}

static int mstrcmp(char * mask,char * adr)
{
  for (;;mask++,adr++)
  { if (*mask=='*') return 0;
    if ((*mask=='?') && (*adr))
      continue;
    if ((*mask==0) && (*adr==0))
      return 0;
    if (toupper(*mask)!=toupper(*adr))
      return 1;
  }
}

int wildmatch(char * addr,char * mask)
{
  if (*mask!='*')
  { if (mstrcmp(mask,addr))
      return 1;
    while ((*mask!='*') && (*mask!=0))
      mask++,addr++;
    if (*mask==0)
      return 0;
  }
  mask++;
  for (;;)
  {
    while (mstrcmp(mask,addr))
    { if (*addr==0)
        return 1;
      addr++;
    }
    while ((*mask!='*') && (*mask!=0))
      mask++,addr++;
    if (*mask==0)
      return 0;
    mask++;
  }
}

void printftnaddr(Address *addr)
{
  if (addr->point)
    printf("%u:%u/%u.%u", addr->zone, addr->net, addr->node, addr->point);
  else
    printf("%u:%u/%u", addr->zone, addr->net, addr->node);
}

int getfidomask(char *addr, Address *node)
/* 頥 0  砥 ᯥ */
{
  char *p, *p1, *p2=addr;
  char c;

  while (isdigit((unsigned char)*p2)||(*p2==':')||(*p2=='/')||(*p2=='.')||(*p2=='*'))
    p2++;
  if (p2==addr) return 1;
  c=*p2;
  *p2=0;
  p=strchr(addr,'*');
  if (p)
  { *p2=c;
    p2=p+1;
    c=*p2;
    *p2=0;
  }
  node->zone=myaka->main.zone;
  node->net=node->node=node->point=(word)-1;
  p=strchr(addr,':');
  if (p)
  { node->zone=atoi(addr);
    addr=p+1;
  }
  else if (addr[0]=='*')
  { node->zone=(word)-1;
    *p2=c;
    return 0;
  }
  p1=strchr(addr,'/');
  if (addr[0]=='*')
  { *p2=c;
    return 0;
  }
  if (p1==addr)
  { node->net=myaka->main.net;
    addr++;
    goto getnode;
  }
  if (p1==NULL)
  { /* may be regional mask */
    if (strchr(addr,'.') || (strchr(addr, '*')==NULL))
    { node->net=myaka->main.net;
      goto getnode;
    }
    node->net=atoi(addr);
    *p2=c;
    if (node->net>=100)
      return 1;
    node->node=(word)-2;
    return 0;
  }
  node->net=atoi(addr);
  addr=p1+1;
getnode:
  if (addr[0]=='*')
  { *p2=c;
    return 0;
  }
  if (addr[0]=='.')
    node->node=myaka->main.node;
  else
    node->node=atoi(addr);
  p=strchr(addr,'.');
  if (p)
  { if (p[1]!='*')
      node->point=atoi(p+1);
  }
  else
    node->point=0; /* 2:463/68 means 2:463/68.0, not 2:463/68.* */
  *p2=c;
  return 0;
}

int getfidoaddr(char *addr, Address *node)
/* 頥 0  砥 ᯥ */
{
  char *p1, *p2=addr;
  char c;

  while (isdigit((unsigned char)*p2)||(*p2==':')||(*p2=='/')||(*p2=='.')||(*p2=='*'))
    p2++;
  if (p2==addr) return 1;
  c=*p2;
  *p2=0;
  node->zone=myaka->main.zone;
  node->net=myaka->main.net;
  node->node=myaka->main.node;
  node->point=0;
  p1=strchr(addr,':');
  if (p1)
  { node->zone=atoi(addr);
    addr=p1+1;
  }
  p1=strchr(addr,'/');
  if (p1==addr)
    addr++;
  else if (p1!=NULL)
  { node->net=atoi(addr);
    addr=p1+1;
  }
  p1=strchr(addr, '.');
  if (p1!=addr)
    node->node=atoi(addr);
  if (p1)
    node->point=atoi(p1+1);
  *p2=c;
  return 0;
}

int checkmask(Address *addr, Address *mask)
{ word n;

  if (mask->zone==(word)-1) return 1;
  if (addr->zone!=mask->zone) return 0;
  if (mask->net==(word)-1) return 1;
  if ((mask->net<100) && (mask->node==(word)-2))
  { /* regional mask */
    for (n=addr->net; n>=100; n/=10);
    if (n==mask->net) return 1;
    return 0;
  }
  if (addr->net!=mask->net) return 0;
  if (mask->node==(word)-1) return 1;
  if (addr->node!=mask->node) return 0;
  if (mask->point==(word)-1) return 1;
  if (addr->point!=mask->point) return 0;
  return 1;
}

char *getnext(int argc, char *argv[])
{
  static int n=0;
  static FILE *f=NULL;
  char *p;
  static char str[256];

  while (1)
  { if (f)
    { if (fgets(str, sizeof(str), f)!=NULL)
      { p=strchr(str, '\n');
        if (p) *p='\0';
        if (*str=='\0') continue;
        return str;
      }
      fclose(f);
      f=NULL;
    }
    if (n==argc) return NULL;
    p=argv[n++];
    if (*p!='@') return p;
    if (strcmp(p+1, "-")==0)
    { int h=dup(fileno(stdin));
      if (h!=-1)
        f=fdopen(h, "r");
      else
        f=NULL;
    }
    else
      f=fopen(p+1, "r");
    if (f==NULL)
      fprintf(stderr, "Can't open %s: %s!\n", p+1, strerror(errno));
  }
}

int findnode(Node ** pnode, int *nnode, Address *mask)
{
#ifdef INCORE
  Node *p;

  p= *pnode ? (Node *)((char *)(*pnode)+basecfg->NodeRecSize) : nodes;
  *nnode=((char *)p-(char *)nodes)/basecfg->NodeRecSize;
  for (;*nnode<basecfg->NodeCnt; (*nnode)++, p=(Node *)((char *)p+basecfg->NodeRecSize))
  { if (checkmask(&p->addr, mask))
    { *pnode=p;
      return 0;
    }
  }
  return 1;
#else
  static Node *p=NULL;

  if (p==NULL) p=malloc(basecfg->NodeRecSize);
  if (p==NULL)
  { fprintf(stderr, "Not enough memory!\n");
    return -1;
  }
  if (*pnode!=NULL) (*nnode)++;
  else *nnode=0;
  lseek(hcfg,
        sizeof(CONFIG)+basecfg->offset+(long)nnode[0]*basecfg->NodeRecSize,
        SEEK_SET);
  for (;*nnode<basecfg->NodeCnt; (*nnode)++)
  { read(hcfg, p, basecfg->NodeRecSize);
    if (checkmask(&p->addr, mask))
    { *pnode=p;
      return 0;
    }
  }
  free(p); p=NULL;
  return 1;
#endif
}

static void checknode(Node *node)
{
  int i;
  char g;

  puts("");
  printf("Address: "); printftnaddr(&node->addr); puts("");
  printf("ArcDest: "); printftnaddr(&node->addr); puts("");
  printf("RealNam: %s\n", node->name);
  printf("Passive: %s\n", node->flags.passive ? "Yes" : "No");
  printf("ArcAttr: %s%s\n", node->flags.arc_status & 1 ?
                            (node->flags.arc_status==1 ? "Hold"  : "Unknown") :
                            (node->flags.arc_status==2 ? "Crash" : "None"),
                            node->flags.arc_direct ? "/Direct" : "");
  printf("Passwrd: %s\n", node->areafixpw);
  printf("PktPass: %s\n", node->password);
  printf("MyAka:   ");
  printftnaddr(node->aka>=basecfg->AkaCnt ? &zeroaddr : &myaka[node->aka].main);
  puts("");
  printf("Packer:  %s\n",
         node->flags.packer==0xf ? "PKT" : packers[node->flags.packer].tag);
  printf("SecLevl: %u\n", node->sec_level);
  printf("ArcSize: %uK\n", node->maxarcsize);
  printf("GrAvail: ");
  for (i=0, g='A'; i<32; i++, g++)
  { if (g=='Z'+1) g='1';
    if (GetBam((byte *)&node->groups, i)==0)
      printf("%c", g);
  }
  puts("");
  printf("Areas:\n");
#ifdef INCORE
  for (i=0; i<basecfg->AreaCnt; i++)
  { Area *area=areas+i;
#else
  lseek(hcfg, sizeof(CONFIG)+basecfg->offset+
        (long)basecfg->NodeCnt*basecfg->NodeRecSize, SEEK_SET);
  for (i=0; i<basecfg->AreaCnt; i++)
  { Area curarea, *area=&curarea;
    read(hcfg, area, sizeof(Area));
#endif
    if (GetBam(node->areas, area->conference))
      printf("  %s\n", area->name);
  }
}

int checkuser(int nusers, char *users[])
{
  Address node;
  Node *pnode;
  int r;
  char *user;

  if ((r=readcfg(1))!=0)
    return r;
  while ((user=getnext(nusers, users))!=NULL)
  { if (stricmp(user, "all")==0) user="*";
    if (getfidomask(user, &node))
    { fprintf(stderr, "%s is not valid FTN-address mask!\n", user);
      continue;
    }
    for (pnode=NULL; findnode(&pnode, &r, &node)==0;)
      checknode(pnode);
    if (pnode==NULL)
      printf("No such user %s\n", user);
  }
  return 0;
}  

static void checkarea(Area *area)
{ Node *pnode=NULL;
  int i, n;

  puts("");
  printf("AreaName: %s\n", area->name);
  printf("Descript: %s\n", area->desc);
  printf("AreaType: ");
  switch(area->flags.atype)
  { case AREA_ECHOMAIL: printf("echomail\n"); break;
    case AREA_NETMAIL:  printf("netmail\n"); break;
    case AREA_LOCAL:    printf("local\n"); break;
    case AREA_BADMAILBOARD: printf("badmail\n"); break;
    case AREA_DUPEBOARD:printf("dupes\n"); break;
    default:            printf("unknown\n"); break;
  }
  printf("Storage:  ");
  switch(area->flags.storage)
  { case QBBS:  printf("HMB, board %u\n", area->board); break;
    case FIDO:  printf("MSG, "); break;
    case SQUISH:printf("Squish, "); break;
    case JAM:   printf("JAM, "); break;
    case PASSTHRU: printf("Passthru\n"); break;
    default:    printf("Unknown\n"); break;
  }
  switch(area->flags.storage)
  { case FIDO:
    case SQUISH:
    case JAM:   printf("%s\n", area->path);
  }
  printf("Group:    %c\n", area->info.group+'A'>'Z' ?
                             area->info.group+'A'-'Z'+'0' :
                             area->info.group+'A');
  printf("Passive:  %s\n", area->advflags.passive   ? "Yes" : "No");
  printf("AKA:      ");
  printftnaddr(area->info.aka>=basecfg->AkaCnt ? &zeroaddr : &myaka[area->info.aka].main);
  puts("");
  printf("RdSecure: %u\n", area->read_sec);
  printf("WrSecure: %u\n", area->write_sec);
  printf("Days:     %d\n", area->days);
  printf("RecvDays: %d\n", area->recvdays);
  printf("Messages: %d\n", area->messages);
  printf("AutoAdd:  %s\n", area->advflags.autoadded ? "Yes" : "No");
  printf("TinySeen: %s\n", area->advflags.tinyseen  ? "Yes" : "No");
  printf("CPD:      %s\n", area->advflags.cpd       ? "Yes" : "No");
  printf("KeepSeen: %s\n", area->advflags.keepseen  ? "Yes" : "No");
  printf("Mandator: %s\n", area->advflags.mandatory ? "Yes" : "No");
  printf("HideArea: %s\n", area->advflags.hide      ? "Yes" : "No");
  printf("Manual:   %s\n", !area->advflags.nomanual ? "Yes" : "No");
  printf("Subscribers:");
  i=n=0;
  while (findnode(&pnode, &n, &allnodes)==0)
    if (GetBam(pnode->areas, area->conference))
    { if (i++%6==0)
        printf("\n  ");
      else
        printf(" ");
      printftnaddr(&pnode->addr);
    }
  puts("");
}

int checkgroup(int ngroups, char *groups[])
{
  int r, i;
  char *group;

  if ((r=readcfg(1))!=0)
    return r;
  while ((group=getnext(ngroups, groups))!=NULL)
  { r=0;
    if (stricmp(group, "all")==0) group="*";
    for (i=0; i<basecfg->AreaCnt; i++)
    {
#ifdef INCORE
      if (wildmatch(areas[i].name, group)==0)
      { checkarea(areas+i);
        r=1;
      }
#else
      Area area;
      lseek(hcfg, sizeof(CONFIG)+basecfg->offset+
                  (long)basecfg->NodeCnt*basecfg->NodeRecSize+
                  (long)i*sizeof(Area), SEEK_SET);
      read(hcfg, &area, sizeof(area));
      if (wildmatch(area.name, group)==0)
      { checkarea(&area);
        r=1;
      }
#endif
    }
    if (r==0)
      printf("No such group %s\n", group);
  }
  return 0;
}  

int subscribe(char *user, int ngroups, char *groups[], int subs)
{
  Address addr;
  Node *node=NULL;
  int r, i, nnode=0, waschanges=0;
  char *group;

  if ((r=readcfg(0))!=0)
    return r;
  if (getfidoaddr(user, &addr))
  { rmbusy();
    fprintf(stderr, "%s is not valid ftn-address!\n", user);
    return 1;
  }
  if (findnode(&node, &nnode, &addr))
  { rmbusy();
    fprintf(stderr, "no such user\n");
    return 1;
  }
  while ((group=getnext(ngroups, groups))!=NULL)
  {
    r=0;
#ifdef INCORE
    for (i=0; i<basecfg->AreaCnt; i++)
    { Area *parea=areas+i;
#else
    lseek(hcfg, sizeof(CONFIG)+basecfg->offset+
                (long)basecfg->NodeCnt*basecfg->NodeRecSize, SEEK_SET);
    for (i=0; i<basecfg->AreaCnt; i++)
    { Area area, *parea;
      parea=&area;
      read(hcfg, parea, sizeof(area));
#endif
      if (wildmatch(parea->name, group)==0)
      { r=1;
        if (GetBam(node->areas, parea->conference)==subs)
        { printf("%s already %s %s\n", user,
                 subs ? "subscribed to" : "unsubscribed from", parea->name);
          continue;
        }
        waschanges=1;
        if (subs) AddBam(node->areas, parea->conference);
        else     FreeBam(node->areas, parea->conference);
        printf("%s %s %s\n", user,
               subs ? "subscribed to" : "unsubscribed from", parea->name);
      }
    }
    if (r==0)
      printf("No such area %s\n", group);
  }
  if (!waschanges)
  { rmbusy();
    return 0;
  }
#ifdef INCORE
  return writecfg();
#else
  if ((nhcfg=newcfg())==-1)
    return 1;
  lseek(hcfg, 0, SEEK_SET);
  if (copybytes(hcfg, nhcfg, sizeof(CONFIG)+basecfg->offset+
                (long)nnode*basecfg->NodeRecSize))
    return 1;
  write(nhcfg, node, basecfg->NodeRecSize);
  lseek(hcfg, basecfg->NodeRecSize, SEEK_CUR);
  if (copybytes(hcfg, nhcfg,
      (long)(basecfg->NodeCnt-nnode-1)*basecfg->NodeRecSize+
      (long)basecfg->AreaCnt*sizeof(Area)))
    return 1;
  if (rencfg(nhcfg, hcfg))
    return 1;
  rmbusy();
  return 0;
#endif
}

int setpassive(char *user, int what)
{
  int r;
  Address addr;
  Node *node;

  if ((r=readcfg(0))!=0)
    return r;
  if (getfidoaddr(user, &addr))
  { rmbusy();
    fprintf(stderr, "%s is not valid ftn-address!\n", user);
    return 1;
  }
  if (findnode(&node, &r, &addr))
  { rmbusy();
    fprintf(stderr, "No such user\n");
    return 1;
  }
  if (node->flags.passive == what)
  { printf("%s already %s\n", user, what ? "passive" : "active");
    rmbusy();
    return 0;
  }
  node->flags.passive = what;
#ifdef INCORE
  return writecfg();
#else
  if ((nhcfg=newcfg())==-1)
    return 1;
  lseek(hcfg, 0, SEEK_SET);
  if (copybytes(hcfg, nhcfg, sizeof(CONFIG)+basecfg->offset+
                (long)r*basecfg->NodeRecSize))
    return 1;
  write(nhcfg, node, basecfg->NodeRecSize);
  lseek(hcfg, basecfg->NodeRecSize, SEEK_CUR);
  if (copybytes(hcfg, nhcfg, (long)(basecfg->NodeCnt-r-1)*basecfg->NodeRecSize))
    return 1;
  if (rencfg(nhcfg, hcfg))
    return 1;
  rmbusy();
  return 0;
#endif  
}

static void charea(Area *area, int argc, char *argv[], int *waschanges)
{
  int  i, j;
  char c, prev=1;
  byte group;

  for (i=0; i<argc; i++)
    for (j=0; j<strlen(argv[i]); j++)
    { c=argv[i][j];
      if (isspace((unsigned char)c)) continue;
      c=toupper(c);
      if (prev=='G')
      { /* set group */
        prev=1;
        if ((group=readgroup(c))==(byte)-1)
        { fprintf(stderr, "Incorrect group %c ignored\n", c);
          continue;
        }
        if (area->info.group!=group)
        { area->info.group=group;
          *waschanges=1;
        }
        continue;
      }
      switch(c)
      {
        case '!':
          if (prev=='!') fprintf(stderr, "Extra '!' ignored\n");
          prev=0;
          continue;
        case 'M': /* mandatory */
          if (area->advflags.mandatory==prev) break;
          area->advflags.mandatory=prev;
          *waschanges=1;
          break;
        case 'S': /* keep seen-byes */
          if (area->advflags.keepseen==prev) break;
          area->advflags.keepseen=prev;
          *waschanges=1;
          break;
        case 'T': /* tiny seen-byes */
          if (area->advflags.tinyseen==prev) break;
          area->advflags.tinyseen=prev;
          *waschanges=1;
          break;
        case 'C': /* CPD */
          if (area->advflags.cpd==prev) break;
          area->advflags.cpd=prev;
          *waschanges=1;
          break;
        case 'R': /* remote changes */
          prev = prev ^ 1; /* "remote changes yes" means "nomanual no" */
          if (area->advflags.nomanual==prev) break;
          area->advflags.nomanual=prev;
          *waschanges=1;
          break;
        case 'H': /* hide area */
          if (area->advflags.hide==prev) break;
          area->advflags.hide=prev;
          *waschanges=1;
          break;
        case 'P': /* passive */
          if (area->advflags.passive==prev) break;
          area->advflags.passive=prev;
          *waschanges=1;
          break;
        case 'G':
          prev=c;
          continue;
        default:
          fprintf(stderr, "Unknown group flag '%c' ignored\n", c);
          break;
      }
      prev=1;
    }
}

int chgroup(char *mask, int argc, char *argv[])
{
  int i, r, waschanges=0;

  if ((r=readcfg(0))!=0)
    return r;
#ifdef INCORE
  for (i=0; i<basecfg->AreaCnt; i++)
    if (wildmatch(areas[i].name, mask)==0)
    { r=1;
      charea(areas+i, argc, argv, &waschanges);
    }
#else
  nhcfg=-1;
  lseek(hcfg, sizeof(CONFIG)+basecfg->offset+
        (long)basecfg->NodeCnt*basecfg->NodeRecSize, SEEK_SET);
  for (i=0; i<basecfg->AreaCnt; i++)
  { Area area;
    read(hcfg, &area, sizeof(Area));
    if (wildmatch(area.name, mask)==0)
    { r=1;
      charea(&area, argc, argv, &waschanges);
      if (waschanges && (nhcfg==-1))
      { if ((nhcfg=newcfg())==-1)
          return 1;
        lseek(hcfg, 0, SEEK_SET);
        if (copybytes(hcfg, nhcfg, sizeof(CONFIG)+basecfg->offset+
            (long)basecfg->NodeCnt*basecfg->NodeRecSize+
            (long)i*sizeof(Area)))
          return 1;
      }
    }
    if (nhcfg!=-1)
      write(nhcfg, &area, sizeof(area));
  }
#endif
  if (r==0)
  { fprintf(stderr, "No such group\n");
    rmbusy();
    return 1;
  }
  if (!waschanges)
  { rmbusy();
    printf("No changes made\n");
    return 0;
  }
#ifdef INCORE
  return writecfg();
#else
  if (copybytes(hcfg, nhcfg, (long)basecfg->NodeCnt*basecfg->NodeRecSize))
    return 1;
  if (rencfg(nhcfg, hcfg))
    return 1;
  rmbusy();
  return 0;
#endif
}

int chuser(char *user, int argc, char *argv[])
{
  int i, nnode=0, r, waschanges=0;
  Address addr;
  Node *node;

  if ((r=readcfg(0))!=0)
    return r;
  if (getfidoaddr(user, &addr))
  { rmbusy();
    fprintf(stderr, "%s is not valid ftn-address!\n", user);
    return 1;
  }
  if (findnode(&node, &nnode, &addr))
  { rmbusy();
    fprintf(stderr, "No such user\n");
    return 1;
  }
  for (i=0; i<argc; i++)
  { strlwr(argv[i]);
    if (i==argc-1)
    { fprintf(stderr, "Parameter %s ignored\n", argv[i]);
      continue;
    }
    if (stricmp(argv[i], "pwd")==0)
    { i++;
      if (strlen(argv[i])>8)
      { fprintf(stderr, "Password truncated to eight chars\n");
        argv[i][8]=0;
      }
      strupr(argv[i]);
      if (strcmp(argv[i], node->areafixpw))
      { strcpy(node->areafixpw, argv[i]);
        waschanges=1;
      }
      continue;
    }
    if (stricmp(argv[i], "pktpwd")==0)
    { i++;
      if (strlen(argv[i])>8)
      { fprintf(stderr, "Pkt password truncated to eight chars\n");
        argv[i][8]=0;
      }
      strupr(argv[i]);
      if (strcmp(argv[i], node->password))
      { strcpy(node->password, argv[i]);
        waschanges=1;
      }
      continue;
    }
    if (stricmp(argv[i], "level")==0)
    { i++;
      if (!isdigit((unsigned char)argv[i][0]))
      { fprintf(stderr, "Incorrect level value '%s' ignored\n", argv[i]);
        continue;
      }
      if (node->sec_level!=(word)atoi(argv[i]))
      { node->sec_level=(word)atoi(argv[i]);
        waschanges=1;
      }
      continue;
    }
    if (stricmp(argv[i], "arcsize")==0)
    { i++;
      if (!isdigit((unsigned char)argv[i][0]))
      { fprintf(stderr, "Incorrect arcsize value '%s' ignored\n", argv[i]);
        continue;
      }
      if (node->maxarcsize!=(word)atoi(argv[i]))
      { node->maxarcsize=(word)atoi(argv[i]);
        waschanges=1;
      }
      continue;
    }
    if (stricmp(argv[i], "attr")==0)
    { byte status, direct;

      i++;
      if (readattr(argv[i], &status, &direct))
      { fprintf(stderr, "Unknown attributes '%s', attributes not changed\n", argv[i]);
        continue;
      }
      if (node->flags.arc_status!=status ||
          node->flags.arc_direct!=direct)
      { node->flags.arc_status=status;
        node->flags.arc_direct=direct;
        waschanges=1;
      }
      continue;
    }
    if (stricmp(argv[i], "groups")==0)
    { dword groups=0;

      i++;
      groups=readgroups(argv[i]);
      if (node->groups!=groups)
      { node->groups=groups;
        waschanges=1;
      }
      continue;
    }
    fprintf(stderr, "Unknown parameter %s ignored\n", argv[i]);
  }
  if (!waschanges)
  { rmbusy();
    printf("No changes made\n");
    return 0;
  }
#ifdef INCORE
  return writecfg();
#else
  if ((nhcfg=newcfg())==-1)
    return 1;
  lseek(hcfg, 0, SEEK_SET);
  if (copybytes(hcfg, nhcfg, sizeof(CONFIG)+basecfg->offset+
                (long)nnode*basecfg->NodeRecSize))
    return 1;
  write(nhcfg, node, basecfg->NodeRecSize);
  lseek(hcfg, basecfg->NodeRecSize, SEEK_CUR);
  if (copybytes(hcfg, nhcfg, (long)(basecfg->NodeCnt-nnode-1)*basecfg->NodeRecSize))
    return 1;
  if (rencfg(nhcfg, hcfg))
    return 1;
  rmbusy();
  return 0;
#endif
}

#ifdef INCORE
static int rmarea(int areanum)
{
  Node *pnode=NULL;
  int nnode=0;

  if (areas[areanum].flags.atype==AREA_BADMAILBOARD ||
      areas[areanum].flags.atype==AREA_DUPEBOARD)
  { printf("Can't delete %s, it's %s\n", areas[areanum].name,
     areas[areanum].flags.atype==AREA_DUPEBOARD ? "dupeboard" : "badmailboard");
    return 1;
  }
  printf("Area %s removed\n", areas[areanum].name);
  /* Unsubscribe all nodes from the area */
  while (findnode(&pnode, &nnode, &allnodes)==0)
    FreeBam(pnode->areas, areas[areanum].conference);
  /* Change AreaCnt */
  basecfg->AreaCnt--;
  /* move areas */
  memcpy(areas+areanum, areas+areanum+1,
         sizeof(*areas)*(basecfg->AreaCnt-areanum));
  return 0;
}

int rmgroup(int nmasks, char *masks[])
{
  int i, r, waschanges=0;
  char *mask;

  if ((r=readcfg(0))!=0)
    return r;
  while ((mask=getnext(nmasks, masks))!=NULL)
  { r=0;
    if (stricmp(mask, "all")==0) mask="*";
    for (i=0; i<basecfg->AreaCnt; i++)
      if (wildmatch(areas[i].name, mask)==0)
      { r++;
        if (rmarea(i)==0)
        { waschanges=1;
          i--;
        }
      }
    if (r==0)
      fprintf(stderr, "No such group %s\n", mask);
  }
  if (!waschanges)
  { rmbusy();
    printf("No changes made\n");
    return 0;
  }
  return writecfg();
}

#else

int rmgroup(int nmasks, char *masks[])
{
  Area area;
  Node *node;
  int  i, r;
  word NewAreaCnt;
  byte rmareas[4096/8]={0}, rmconfs[4096/8]={0};
  char *mask;

  if ((r=readcfg(0))!=0)
    return r;
  NewAreaCnt=basecfg->AreaCnt;
  while ((mask=getnext(nmasks, masks))!=NULL)
  { r=0;
    if (stricmp(mask, "all")==0) mask="*";
    lseek(hcfg, sizeof(CONFIG)+basecfg->offset+
          (long)basecfg->NodeCnt*basecfg->NodeRecSize, SEEK_SET);
    for (i=0; i<basecfg->AreaCnt; i++)
    { read(hcfg, &area, sizeof(Area));
      if (wildmatch(area.name, mask)==0)
      { if (area.flags.atype==AREA_BADMAILBOARD ||
            area.flags.atype==AREA_DUPEBOARD)
        { printf("Can't delete %s, it's %s\n", area.name,
               area.flags.atype==AREA_DUPEBOARD ? "dupeboard" : "badmailboard");
          continue;
        }
        AddBam(rmareas, i);
        AddBam(rmconfs, area.conference);
        NewAreaCnt--;
        r++;
      }
    }
    if (r==0)
      printf("No such group %s\n", mask);
  }
  if (NewAreaCnt==basecfg->AreaCnt)
  { printf("No changes made\n");
    rmbusy();
    return 0;
  }
  node=malloc(basecfg->NodeRecSize);
  if (node==NULL)
  { fprintf(stderr, "Not enough memory!\n");
    rmbusy();
    return 1;
  }
  if ((nhcfg=newcfg())==-1)
    return 1;
  i=basecfg->AreaCnt;
  basecfg->AreaCnt=NewAreaCnt;
  write(nhcfg, basecfg, sizeof(CONFIG));
  NewAreaCnt=(word)i;
  lseek(hcfg, sizeof(CONFIG), SEEK_SET);
  copybytes(hcfg, nhcfg, basecfg->offset);
  for (i=0; i<basecfg->NodeCnt; i++)
  { read(hcfg, node, basecfg->NodeRecSize);
    for (r=0; r<basecfg->MaxAreas/8; r++)
      node->areas[r]&=rmconfs[r]^0xff;
    write(nhcfg, node, basecfg->NodeRecSize);
  }
  for (i=0; i<NewAreaCnt; i++)
  { read(hcfg, &area, sizeof(Area));
    if (!GetBam(rmareas, i))
    { if (write(nhcfg, &area, sizeof(Area))!=sizeof(Area))
      { fprintf(stderr, "Can't write to %s: %s!\n", newcfgname, strerror(errno));
        close(nhcfg);
        unlink(newcfgname);
        rmbusy();
        return 1;
      }
    }
    else
      printf("Area %s removed\n", area.name);
  }
  if (rencfg(nhcfg, hcfg))
    return 1;
  rmbusy();
  return 0;
}
#endif

#ifdef INCORE
static void rmnode(Node *node)
{ int i;

  basecfg->NodeCnt--;
  i=((char *)node-(char *)nodes)/basecfg->NodeRecSize;
  memcpy(node, (char *)node+basecfg->NodeRecSize,
         basecfg->NodeRecSize*(basecfg->NodeCnt-i)+
         sizeof(Area)*basecfg->AreaCnt);
}

#endif

int rmuser(int nusers, char *users[])
{
  Address node;
  Node *pnode;
  int r, nnode, waschanges=0;
  char *user;
#ifndef INCORE
  byte rmusers[1024/8]={0};
#endif

  if ((r=readcfg(0))!=0)
    return r;
  while ((user=getnext(nusers, users))!=NULL)
  { r=0;
    if (stricmp(user, "all")==0) user="*";
    if (getfidomask(user, &node))
    { fprintf(stderr, "%s is not valid FTN-address mask!\n", user);
      continue;
    }
    nnode=0;
    for (pnode=NULL; findnode(&pnode, &nnode, &node)==0;)
    { r++;
      waschanges++;
#ifdef INCORE
      rmnode(pnode);
      pnode=NULL; /* from start */
#else
      AddBam(rmusers, nnode);
#endif
    }
    if (r==0)
      printf("No such user %s!\n", user);
  }
  if (!waschanges)
  { printf("No changes made\n");
    rmbusy();
    return 0;
  }
#ifdef INCORE
  return writecfg();
#else
  pnode=malloc(basecfg->NodeRecSize);
  if (pnode==NULL)
  { fprintf(stderr, "Not enough memory!\n");
    return 1;
  }
  nnode=basecfg->NodeCnt;
  basecfg->NodeCnt-=waschanges;
  if ((nhcfg=newcfg())==-1)
    return 1;
  write(nhcfg, basecfg, sizeof(CONFIG));
  lseek(hcfg, sizeof(CONFIG), SEEK_SET);
  if (copybytes(hcfg, nhcfg, basecfg->offset))
    return 1;
  basecfg->NodeCnt=nnode;
  for (nnode=0; nnode<basecfg->NodeCnt; nnode++)
  { read(hcfg, pnode, basecfg->NodeRecSize);
    if (!GetBam(rmusers, nnode))
      if (write(nhcfg, pnode, basecfg->NodeRecSize)!=basecfg->NodeRecSize)
      { fprintf(stderr, "Can't write to %s: %s!\n", newcfgname, strerror(errno));
        close(nhcfg);
        unlink(newcfgname);
        return 1;
      }
  }
  if (copybytes(hcfg, nhcfg, (long)basecfg->AreaCnt*sizeof(Area)))
    return 1;
  if (rencfg(nhcfg, hcfg))
    return 1;
  rmbusy();
  return 0;
#endif
}

int addgroup(char *name, char *desc, char *sgroup, char *stype, char *sstorage,
             char *path)
{
  int  r, i;
  byte group, storage, type;
  word board=NO_BOARD;
  byte boards[256/8]={0};
  byte confer[4096/8]={0};
  Area *area;
  GroupDefaults *gd;
#ifdef INCORE
  long lcfg;
  char *newcfg;
#else
  Area newarea;
#endif

  strupr(name);
  if (strlen(name)>=sizeof(area->name))
  { printf("Area name truncated to %d chars\n", (int)sizeof(area->name)-1);
    name[sizeof(area->name)-1]='\0';
  }
  if (strlen(desc)>=sizeof(area->desc))
  { printf("Area description truncated to %d chars\n", (int)sizeof(area->desc)-1);
    desc[sizeof(area->desc)-1]='\0';
  }
  strupr(sgroup);
  if (strlen(sgroup)!=1)
    printf("Area group must be a single character, I'll use '%c'\n", sgroup[0]);
  if ((group=readgroup(sgroup[0]))==(byte)-1)
  { fprintf(stderr, "Incorrect area group '%c'!\n", sgroup[0]);
    return 1;
  }
  if (stricmp(stype, "echomail")==0)
    type=AREA_ECHOMAIL;
  else if (stricmp(stype, "local")==0)
    type=AREA_LOCAL;
  else if (stricmp(stype, "netmail")==0)
    type=AREA_NETMAIL;
  else
  { fprintf(stderr, "Unknown area type %s\n", stype);
    fprintf(stderr, "Type must be \"echomail\", \"netmail\" or \"local\"\n");
    return 1;
  }
  if (stricmp(sstorage, "passthru")==0)
  { storage=PASSTHRU;
    path="";
  }
  else if (stricmp(sstorage, "HMB")==0 ||
           stricmp(sstorage, "QBBS")==0)
  { if (path && *path)
    { if (!isdigit((unsigned char)*path))
      {
badstorage:
        fprintf(stderr, "Incorrect storage: \"%s%s%s\"\n", sstorage,
                path ? " " : "", path ? path : "");
        fprintf(stderr, "Valid values:\n");
        fprintf(stderr, "  Passthru\n");
        fprintf(stderr, "  HMB <board>\n");
        fprintf(stderr, "  MSG <path>\n");
        fprintf(stderr, "  JAM <path>\n");
        fprintf(stderr, "  SQUISH <path>\n");
        return 1;
      }
      board=atoi(path);
      if (board<1 || board>200)
      { fprintf(stderr, "Bad HMB board value %s, must be 1..200\n", path);
        return 1;
      }
    }
    path="";
    storage=QBBS;
  }
  else
  { char *p, *p1;

    if (path==NULL || *path=='\0') goto badstorage;
    if (stricmp(sstorage, "JAM")==0)
      storage=JAM;
    else if (stricmp(sstorage, "MSG")==0)
      storage=FIDO;
    else if (stricmp(sstorage, "SQUISH")==0)
      storage=SQUISH;
    else
      goto badstorage;

    if (strlen(path)>=sizeof(area->path))
    { fprintf(stderr, "Path too long!\n");
      return 1;
    }
    strupr(path);
    p=strdup(path);
    if (p==NULL)
    { fprintf(stderr, "Not enough memory!\n");
      return 1;
    }
    if (storage!=FIDO)
    { p1=strrchr(p, '\\');
      if (p1) *p1='\0';
    }
    else
      if (p[strlen(p)-1]=='\\')
	p[strlen(p)-1]='\0';
    for (p1=p; (p1=strchr(p1+1, '\\'))!=NULL;)
    {
      *p1='\0';
      mkdir(p);
      *p1='\\';
    }
    mkdir(p);
    free(p);
  }

  if ((r=readcfg(0))!=0)
    return r;
  if (basecfg->AreaCnt==basecfg->MaxAreas)
  { rmbusy();
    fprintf(stderr, "Can't add area: max area count exceeded!\n");
    fprintf(stderr, "Use \"fesetup -R<n>,<a>\" to increase maximum area records\n");
    return 1;
  }
#ifdef INCORE
  lcfg=sizeof(*basecfg)+basecfg->offset+
       basecfg->NodeCnt*basecfg->NodeRecSize+
       basecfg->AreaCnt*basecfg->AreaRecSize;
  newcfg=realloc(cfg, lcfg+sizeof(Area));
  if (newcfg==NULL)
  { rmbusy();
    fprintf(stderr, "Not enough memory!\n");
    return 1;
  }
  basecfg=(CONFIG *)((char *)basecfg+(newcfg-cfg));
  nodes=(Node *)((char *)nodes+(newcfg-cfg));
  areas=(Area *)((char *)areas+(newcfg-cfg));
  myaka=(SysAddress *)((char *)myaka+(newcfg-cfg));
  packers=(Packers *)((char *)packers+(newcfg-cfg));
  cfg=newcfg;
  area=areas+basecfg->AreaCnt;
#else
  area=&newarea;
#endif

  memset(area, 0, sizeof(Area));
  /* set defaults */
  for (gd=grpdefs, i=0; i<basecfg->GDCnt;
       i++, gd=(GroupDefaults *)((char *)gd+basecfg->GrpDefRecSize))
    if (gd->group==group)
    { memcpy(area, &gd->area, sizeof(Area));
      break;
    }
  strcpy(area->name, name);
  strcpy(area->desc, desc);
  strcpy(area->path, path);
  area->board=board;
  area->flags.atype=type;
  area->flags.storage=storage;
  area->info.group=group;
  /* get free conference */
#ifdef INCORE
  for (i=0; i<basecfg->AreaCnt; i++)
  { Area *parea=areas+i;
#else
  lseek(hcfg, sizeof(CONFIG)+basecfg->offset+
        (long)basecfg->NodeCnt*basecfg->NodeRecSize, SEEK_SET);
  for (i=0; i<basecfg->AreaCnt; i++)
  { Area curarea, *parea;
    read(hcfg, &curarea, sizeof(Area));
    parea=&curarea;
#endif
    if (type==QBBS && board==parea->board)
    { rmbusy();
      fprintf(stderr, "HMB board %d already used by area %s\n",
              board, parea->name);
      return 1;
    }
    AddBam(confer, parea->conference);
    if (type==QBBS && board==NO_BOARD)
      if (parea->flags.atype==QBBS && parea->board<256)
        AddBam(boards, parea->board);
  }
  for (i=0; i<basecfg->MaxAreas; i++)
    if (!GetBam(confer, i))
      break;
  area->conference=(word)i;
  if (type==QBBS && board==NO_BOARD)
  { for (i=0; i<200; i++)
      if (!GetBam(boards, i))
        break;
    if (i==200)
    { rmbusy();
      fprintf(stderr, "No free HMB board available!\n");
      return 1;
    }
    area->board=i;
  }
  basecfg->AreaCnt++;
#ifdef INCORE
  return writecfg();
#else
  if ((nhcfg=newcfg())==-1)
    return 1;
  write(nhcfg, basecfg, sizeof(CONFIG));
  lseek(hcfg, sizeof(CONFIG), SEEK_SET);
  if (copybytes(hcfg, nhcfg, basecfg->offset+
      (long)basecfg->NodeCnt*basecfg->NodeRecSize+
      (long)(basecfg->AreaCnt-1)*sizeof(Area)))
    return 1;
  if (write(nhcfg, area, sizeof(Area))!=sizeof(Area))
  { fprintf(stderr, "Can't write to %s: %s!\n", newcfgname, strerror(errno));
    close(nhcfg);
    unlink(newcfgname);
    rmbusy();
    return 1;
  }
  if (rencfg(nhcfg, hcfg))
    return 1;
  rmbusy();
  return 0;
#endif
}

int adduser(char *user, char *sysop, char *passwd, char *attr, char *sgroups,
            char *ssecurity, char *pktpwd)
{
  Address addr;
  dword groups=0;
  byte status=0, direct=0;
  int  r;
  word security;
  Node *node;
#ifdef INCORE
  long i;
  long lcfg;
  char *newcfg;
#else
  Node newnode;
#endif

  if (!getfidoaddr(user, &addr))
  { fprintf(stderr, "%s is not valid FTN-address!\n", user);
    return 1;
  }
  if (strlen(sysop)>=sizeof(node->name))
  { printf("Sysop name truncated to %d chars\n", (int)sizeof(node->name)-1);
    sysop[sizeof(node->name)-1]='\0';
  }
  strupr(passwd);
  if (strlen(passwd)>=sizeof(node->areafixpw))
  { printf("Password truncated to %d chars\n", (int)sizeof(node->areafixpw)-1);
    passwd[sizeof(node->areafixpw)-1]='\0';
  }
  strupr(pktpwd);
  if (strlen(pktpwd)>=sizeof(node->password))
  { printf("PKT password truncated to %d chars\n", (int)sizeof(node->password)-1);
    pktpwd[sizeof(node->password)-1]='\0';
  }
  if (readattr(attr, &status, &direct))
  { fprintf(stderr, "Incorrect attributes '%s'!\n", attr);
    return 1;
  }
  groups=readgroups(sgroups);
  if (!isdigit((unsigned char)ssecurity[0]))
  { fprintf(stderr, "Incorrect security level %s!\n", ssecurity);
    return 1;
  }
  security=(word)atoi(ssecurity);

  if ((r=readcfg(0))!=0)
    return r;
  if (basecfg->NodeCnt==basecfg->MaxNodes)
  { rmbusy();
    fprintf(stderr, "Can't add node: max node count exceeded!\n");
    fprintf(stderr, "Use \"fesetup -R<n>,<a>\" to increase maximum node records\n");
    return 1;
  }
#ifdef INCORE
  lcfg=sizeof(*basecfg)+basecfg->offset+
       (long)basecfg->NodeCnt*basecfg->NodeRecSize+
       (long)basecfg->AreaCnt*basecfg->AreaRecSize;
  newcfg=realloc(cfg, lcfg+basecfg->NodeRecSize);
  if (newcfg==NULL)
  { rmbusy();
    fprintf(stderr, "Not enough memory!\n");
    return 1;
  }
  basecfg=(CONFIG *)((char *)basecfg+(newcfg-cfg));
  nodes=(Node *)((char *)nodes+(newcfg-cfg));
  areas=(Area *)((char *)areas+(newcfg-cfg));
  myaka=(SysAddress *)((char *)myaka+(newcfg-cfg));
  packers=(Packers *)((char *)packers+(newcfg-cfg));
  cfg=newcfg;

  for (i=lcfg-1; i>=(char *)areas-cfg; i++)
    cfg[i+basecfg->NodeRecSize]=cfg[i];
  areas=(Area *)((char *)areas+basecfg->NodeRecSize);

  node=(Node *)((char *)nodes+basecfg->NodeRecSize*basecfg->NodeCnt);
#else
  node=&newnode;
#endif
  memset(node, 0, basecfg->NodeRecSize);
  strcpy(node->name, sysop);
  memcpy(&node->addr, &addr, sizeof(addr));
  memcpy(&node->arcdest, &addr, sizeof(addr));
  strcpy(node->password, pktpwd);
  strcpy(node->areafixpw, passwd);
  node->sec_level=security;
  node->groups=groups;
  node->flags.arc_status=status;
  node->flags.arc_direct=direct;
  /* set defaults */
  node->flags.dddd=1;
  basecfg->NodeCnt++;
#ifdef INCORE
  return writecfg();
#else
  if ((nhcfg=newcfg())==-1)
    return 1;
  write(nhcfg, basecfg, sizeof(CONFIG));
  lseek(hcfg, sizeof(CONFIG), SEEK_SET);
  if (copybytes(hcfg, nhcfg, basecfg->offset+
      (long)(basecfg->NodeCnt-1)*basecfg->NodeRecSize))
    return 1;
  if (write(nhcfg, node, basecfg->NodeRecSize)!=basecfg->NodeRecSize)
  { fprintf(stderr, "Can't write to %s: %s!\n", newcfgname, strerror(errno));
    close(nhcfg);
    unlink(newcfgname);
    rmbusy();
    return 1;
  }
  if (copybytes(hcfg, nhcfg, (long)basecfg->AreaCnt*sizeof(Area)))
    return 1;
  if (rencfg(nhcfg, hcfg))
    return 1;
  rmbusy();
  return 0;
#endif
}

struct {
  char *name;
  int  offs;
  word bits;
  word value;
} groupprop[] = {
  { "tinyseen", 64, 0x0002, 0x0002 },
  { "cpd",      64, 0x0004, 0x0004 },
  { "passive",  64, 0x0008, 0x0008 },
  { "keepseen", 64, 0x0010, 0x0010 },
  { "mandatory",64, 0x0020, 0x0020 },
  { "hide",     64, 0x0400, 0x0400 },
  { "remote",   64, 0x0800, 0x0800 },
  { "echomail", 62, 0x00f0, AREA_ECHOMAIL<<4 },
  { "local",    62, 0x00f0, AREA_LOCAL   <<4 },
  { "netmail",  62, 0x00f0, AREA_NETMAIL <<4 },
  { "badmail",  62, 0x00f0, AREA_BADMAILBOARD <<4 },
  { "dupemail", 62, 0x00f0, AREA_DUPEBOARD <<4 },
  { "jam",      62, 0x000f, JAM  },
  { "msg",      62, 0x000f, FIDO },
  { "hmb",      62, 0x000f, QBBS },
  { "qbbs",     62, 0x000f, QBBS },
  { "squish",   62, 0x000f, SQUISH },
  { "passthru", 62, 0x000f, PASSTHRU }
};

#define GROUPPARAMS sizeof(groupprop)/sizeof(*groupprop)

int listgroups(char *group, int argc, char *argv[])
{
  int i, j, r;
  char *p;
  Area *area;
  dword groups=0xffffffffl, nogroups=0;
#ifndef INCORE
  Area curarea;
#endif

  if (stricmp(group, "all")==0) group="*";
  /* check syntax of argv */
  for (r=0; r<argc; r++)
  { p=argv[r];
    if (*p=='!') p++;
    for (i=0; i<GROUPPARAMS; i++)
      if (stricmp(p, groupprop[i].name)==0)
        break;
    if (i<GROUPPARAMS) continue;
    if (strnicmp(p, "groups:", 7)==0)
    { if (argv[r][0]=='!')
        nogroups=readgroups(p+7);
      else
        groups=readgroups(p+7);
    }
    else
      fprintf(stderr, "Unknown keyword %s ignored\n", p);
  }
  if ((r=readcfg(1))!=0)
    return r;
#ifdef INCORE
  for (i=0; i<basecfg->AreaCnt; i++)
  { area=areas+i;
#else
  area=&curarea;
  lseek(hcfg, sizeof(CONFIG)+basecfg->offset+
        (long)basecfg->NodeCnt*basecfg->NodeRecSize, SEEK_SET);
  for (i=0; i<basecfg->AreaCnt; i++)
  { read(hcfg, &curarea, sizeof(Area));
#endif
    if (wildmatch(area->name, group))
      continue;
    if (GetBam((byte *)&nogroups, area->info.group))
      continue;
    if (!GetBam((byte *)&groups, area->info.group))
      continue;
    for (j=0; j<argc; j++)
    { p=argv[j];
      if (*p=='!') p++;
      for (r=0; r<GROUPPARAMS; r++)
      { if (stricmp(p, groupprop[r].name)==0)
        { word bitflags=*(word *)((char *)area+groupprop[r].offs);
          if (((bitflags & groupprop[r].bits) == groupprop[r].value) ==
              (argv[j][0]=='!'))
          { r=-1; /* not matched */
            break;
          }
        }
      }
      if (r==-1) break;
    }
    if (j==argc)
      printf("%s\n", area->name);
  }
  rmbusy();
  return 0;
}

void help(void)
{
  puts("FESDBA - change fastecho 1.46 cfg in packet mode");
  puts("Version 0.1          " __DATE__);
  puts("Copyright (C) Pavel Gulchouck 2:463/68   gul@lucky.net");
  puts("Freeware");
  puts("\n  Usage:");
  puts("fesdba check node <mask> [<mask> ...]");
  puts("fesdba check area <mask> [<mask> ...]");
  puts("fesdba subscribe <node> <mask> [<mask> ...]");
  puts("fesdba unsubscribe <node> <mask> [<mask> ...]");
  puts("fesdba disable <node>");
  puts("fesdba enable <node>");
  puts("fesdba charea <mask> <flags>");
  puts("fesdba chnode <flags>");
  puts("fesdba rmarea <mask> [<mask> ...]");
  puts("fesdba rmnode <mask> [<mask> ...]");
  puts("fesdba addarea <name> <desc> <group> <type> <storage>");
  puts("fesdba addnode <node> <sysop> <passwd> <attrib> <groups> <security> [<pkt passwd>]");
  puts("fesdba list areas <mask> [[!]<param> ...]");
  puts("");
  puts("@<filename> means get list of masks from file");
  puts("");
  puts("  area flags:  (!<flag> - switch off)");
  puts("M - mandatory,");
  puts("S - keep seen-byes");
  puts("T - tiny seen-byes");
  puts("C - CPD");
  puts("R - remote changes");
  puts("H - hide area");
  puts("P - passive");
  puts("G[A-Z] - group");
  puts("");
  puts("  Node flags:");
  puts("pwd <passwd>    - set passwd");
  puts("pktpwd <passwd> - set pkt passwd");
  puts("attr <attr>     - set arcmail attr");
  puts("groups <groups> - set available groups");
  puts("level <level>   - set security level");
  puts("arcsize <size>  - set maximum arcmail size");
}

int main(int argc, char *argv[])
{
  char *myname, *p;

  if (argc<3)
  { help();
    return 0;
  }

  /***************** get cfg name ********************/
  myname=argv[0];
#ifdef __OS2__
  { PPIB pib;
    PTIB tib;
    DosGetInfoBlocks(&tib, &pib);
    if (pib)
    { for (myname=pib->pib_pchenv; myname[0] || myname[1]; myname++);
      myname+=2;
    }
  }
#endif
  strcpy(cfgname, myname);
  p=strrchr(cfgname, '\\');
  if (p==NULL) p=cfgname;
  else p++;
  strcpy(p, "fastecho.cfg");

  /* set signal handlers */
  signal(SIGINT,   cbreak);
  signal(SIGTERM,  cbreak);
#ifdef SIGBREAK
  signal(SIGBREAK, cbreak);
#endif
#ifdef __MSDOS__
  setcbrk(1);
#endif

  /* parse args */
  strlwr(argv[1]);
  if (strcmp(argv[1], "check")==0)
  { if (argc<4)
    { help();
      return 1;
    }
    strlwr(argv[2]);
    if (strcmp(argv[2], "node")==0)
      return checkuser(argc-3, argv+3);
    else if (strcmp(argv[2], "area")==0)
      return checkgroup(argc-3, argv+3);
    help();
    return 1;
  }
  if (strcmp(argv[1], "subscribe")==0)
  { if (argc<4)
    { help();
      return 1;
    }
    return subscribe(argv[2], argc-3, argv+3, 1);
  }
  if (strcmp(argv[1], "unsubscribe")==0)
  { if (argc<4)
    { help();
      return 1;
    }
    return subscribe(argv[2], argc-3, argv+3, 0);
  }
  if (strcmp(argv[1], "disable")==0)
    return setpassive(argv[2], 1);
  if (strcmp(argv[1], "enable")==0)
    return setpassive(argv[2], 0);
  if (strcmp(argv[1], "charea")==0)
    return chgroup(argv[2], argc-3, argv+3);
  if (strcmp(argv[1], "chnode")==0)
    return chuser(argv[2], argc-3, argv+3);
  if (strcmp(argv[1], "rmarea")==0)
    return rmgroup(argc-2, argv+2);
  if (strcmp(argv[1], "rmnode")==0)
    return rmuser(argc-2, argv+2);
  if (strcmp(argv[1], "addarea")==0)
  {
    if (argc!=7)
    { help();
      return 1;
    }
    return addgroup(argv[2], argv[3], argv[4], argv[5], argv[6],
                   argc<8 ? NULL : argv[7]);
  }
  if (strcmp(argv[1], "addnode")==0)
  {
    if (argc<8)
    { help();
      return 1;
    }
    return adduser(argv[2], argv[3], argv[4], argv[5], argv[6], argv[7],
                   argc<8 ? "" : argv[8]);
  }
  if (strcmp(argv[1], "list")==0)
  { if (argc<4)
    { help();
      return 1;
    }
    strlwr(argv[2]);
    if (strcmp(argv[2], "areas")==0)
      return listgroups(argv[3], argc-4, argv+4);
/*
    else if (strcmp(argv[2], "nodes")==0)
      return listusers(argv[3], argc-4, argv+4);
*/
    help();
    return 1;
  }
  help();
  return 1;
}
